{-# OPTIONS -fglasgow-exts #-}

-----------------------------------------------------------------------------
-- |
-- Maintainer  :  joost.visser@di.uminho.pt, alcino@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
-- Point-free Type-preserving Rewriting
--
-----------------------------------------------------------------------------

module Data.Transform.Rewriting where

import Data.Transform.Type

import Control.Monad
import Control.Monad.Reader
import Control.Monad.State
import Control.Monad.RWS
import Data.List

-- SYB reloaded

data Typed a = Type a :| a

data Spine a where
    As :: a -> Con -> Spine a
    Ap :: Spine (a -> b) -> Typed a -> Spine b

data Fixity = Prefix | Infix deriving Eq

data Con = Con {name :: String, fixity :: Fixity}

fromSpine :: Spine a -> a
fromSpine (c `As` _) = c
fromSpine (Ap f (_ :| a)) = (fromSpine f) a

dyncon = Con {name = "Dyn", fixity = Prefix}
intcon n = Con {name = show n, fixity = Prefix}
onecon = Con {name = "_L", fixity = Prefix}
holecon = Con {name = "_", fixity = Prefix}
bangcon = Con {name = "bang", fixity = Prefix}
idcon = Con {name = "id", fixity = Prefix}
compcon = Con {name = ".", fixity = Infix}
fstcon = Con {name = "fst", fixity = Prefix}
sndcon = Con {name = "snd", fixity = Prefix}
splitcon = Con {name = "/\\", fixity = Infix}
prodcon = Con {name = "><", fixity = Infix}
inlcon = Con {name = "inl", fixity = Prefix}
inrcon = Con {name = "inr", fixity = Prefix}
eithercon = Con {name = "\\/", fixity = Infix}
sumcon = Con {name = "-|-", fixity = Infix}
apcon = Con {name = "ap", fixity = Prefix}
currycon = Con {name = "curry", fixity = Prefix}
expcon = Con {name = "exp", fixity = Prefix}
varcon n = Con {name = n, fixity = Prefix}
pntcon = Con{name = "pnt", fixity = Prefix}

nilcon = Con{name="nil", fixity=Prefix}
conscon = Con{name="cons", fixity=Prefix}
listelimcon = Con{name="listelim", fixity=Prefix}
listfoldcon = Con{name="listfold", fixity=Prefix}
listcon = Con{name="list", fixity=Prefix}
list2mapcon = Con{name="list2map", fixity=Prefix}

emptycon = Con{name="empty", fixity=Prefix}
insertcon = Con{name="insert", fixity=Prefix}
mapfoldcon = Con{name="mapfold", fixity=Prefix}
mapfoldccon = Con{name="mapfoldc", fixity=Prefix}
elemscon = Con{name="elems", fixity=Prefix}
ljoincon = Con{name="ljoin", fixity=Prefix}
mapcon = Con{name="map", fixity=Prefix}

somelist = Con{name="[...]",fixity=Prefix}

toSpine :: Type a -> a -> Spine a
toSpine Dynamic (Dyn t x) = Ap ((Dyn t) `As` dyncon) (t :| x)
toSpine Int n = n `As` (intcon n)
toSpine Bool n = n `As` (intcon n)
toSpine Char n = n `As` (intcon n)
toSpine String n = n `As` (intcon n)
toSpine One x = x `As` bangcon
toSpine (List Int) l = l `As` somelist
toSpine (Either a _) (Left x) = Ap (Left `As` Con{name="Left", fixity=Prefix}) (a :| x)
toSpine (Either _ b) (Right y) = Ap (Right `As` Con{name="Left", fixity=Prefix}) (b :| y)
toSpine (PF _) HOLE = HOLE `As` holecon
toSpine (PF _) BANG = BANG `As` bangcon
toSpine (PF (Func a b)) (FUN n f) = (FUN n f) `As` (varcon n)
toSpine (PF (Func one a)) (PNT x) = Ap (PNT `As` pntcon) (a :| x)
toSpine (PF _) ID = ID `As` idcon
toSpine (PF (Func a c)) (COMP b f g) = Ap (Ap ((COMP b) `As` compcon) (PF (Func b c) :| f)) (PF (Func a b) :| g)
toSpine (PF _) FST = FST `As` fstcon
toSpine (PF _) SND = SND `As` sndcon
toSpine (PF (Func a (Prod b c))) (SPLIT f g) = Ap (Ap (SPLIT `As` splitcon) (PF (Func a b) :| f)) (PF (Func a c) :| g)
toSpine (PF (Func (Prod a c) (Prod b d))) (PROD f g) = Ap (Ap (PROD `As` prodcon) (PF (Func a b) :| f)) (PF (Func c d) :| g)
toSpine (PF _) INL = INL `As` inlcon
toSpine (PF _) INR = INR `As` inrcon
toSpine (PF (Func (Either a b) c)) (EITHER f g) = Ap (Ap (EITHER `As` eithercon) (PF (Func a c) :| f)) (PF (Func b c) :| g)
toSpine (PF (Func (Either a c) (Either b d))) (SUM f g) = Ap (Ap (SUM `As` sumcon) (PF (Func a b) :| f)) (PF (Func c d) :| g)
toSpine (PF _) AP = AP `As` apcon
toSpine (PF (Func a (Func b c))) (CURRY f) = Ap (CURRY `As` currycon) (PF (Func (Prod a b) c) :| f) 
toSpine (PF (Func (Func _ a) (Func _ b))) (EXP f) = Ap (EXP `As` expcon) (PF (Func a b) :| f) 
toSpine (PF _) NIL = NIL `As` nilcon 
toSpine (PF _) CONS = CONS `As` conscon 
toSpine (PF (Func (List a) (List b))) (LIST f) = Ap (LIST `As` listcon) (PF (Func a b) :| f)
toSpine (PF _) LIST2MAP = LIST2MAP `As` list2mapcon
toSpine (PF _) EMPTY = EMPTY `As` emptycon 
toSpine (PF _) INSERT = INSERT `As` insertcon 
toSpine (PF _) ELEMS = ELEMS `As` elemscon
toSpine (PF _) LJOIN = LJOIN `As` ljoincon
toSpine (PF _) (SET f) = error $ "toSpine missing case for SET"
toSpine (PF (Func (Map a b) (Map _ c))) (MAP f) = Ap (MAP `As` mapcon) (PF (Func b c) :| f)
toSpine t a = error $ "toSpine missing case for "++show t
    
rmTags :: Type a -> Type a
rmTags (List a) = List (rmTags a)
rmTags (Set a) = Set (rmTags a)
rmTags (Map a b) = Map (rmTags a) (rmTags b)
rmTags (Either a b) = Either (rmTags a) (rmTags b)
rmTags (Prod a b) = Prod (rmTags a) (rmTags b)
rmTags (Func a b) = Func (rmTags a) (rmTags b)
rmTags (Tag _ a) = rmTags a
rmTags (Mu f) = Mu f      -- should descend into f as well...
rmTags (PF a) = PF (rmTags a)
rmTags x = x    

-- Some generic functions

geq :: Type a -> a -> a -> Bool
geq t x y = xyz t t x y
    where xyz :: Type a -> Type b -> a -> b -> Bool
	  xyz t1 t2 x y = aux (toSpine t1 x) (toSpine t2 y)
	  aux :: Spine a -> Spine b -> Bool
	  aux (_ `As` c1)        (_ `As` c2)         = name c1 == name c2
	  aux (Ap f1 (t1 :| a1)) (Ap f2 (t2 :| a2))  = aux f1 f2 && xyz t1 t2 a1 a2
	  aux _ _ = False

gshow :: Type a -> a -> String
gshow Dynamic (Dyn t x) = gshow t x
gshow t x   =  aux (toSpine t x)
    where aux :: Spine a -> String
          aux (Ap (Ap (_ `As` c) (a :| x)) (b :| y)) | fixity c == Infix = "(" ++ gshow a x ++ " " ++ name c ++ " " ++ gshow b y ++ ")"
          aux (_ `As` c)       = name c
          aux (Ap f (t :| a))  =  "(" ++ aux f ++ " " ++ gshow t a ++ ")"

-- Generic queries

type GenericQ r = forall a. Type a -> a -> r

gmapQ :: GenericQ r -> GenericQ [r]
gmapQ q t x = aux q (toSpine t x)
    where aux :: GenericQ r -> (forall a. Spine a -> [r])
          aux q (_ `As` _)      = []
          aux q (Ap f (t :| x)) = aux q f ++ [q t x]

type Query r = forall a . Typed a -> r

everything :: (r -> r -> r) -> GenericQ r -> Query r
everything op q (t :| x) = aux op q t x
    where aux :: (r -> r -> r) -> GenericQ r -> GenericQ r
	  aux op q t x = foldl1 op ([q t x] ++ gmapQ (aux op q) t x)
	       
-- Locations and contexts

type Location = [Int]

down :: Location -> Location
down = (0:)

next :: Location -> Location
next (h:t) = (h+1):t

replace :: Location -> Typed b -> Typed a -> Maybe a
replace [0] (b :| x) (a :| y) = coerce b a x
replace l   (b :| x) (a :| y) = do s <- aux (last l) (init l) (toSpine a y)
                                   return $ fromSpine s
    where aux :: Int -> Location -> Spine a -> Maybe (Spine a)
          aux 0     l (Ap f (a :| y)) = do z <- replace l (b :| x) (a :| y)
                                           return $ Ap f (a :| z)
          aux (n+1) l (Ap f (a :| y)) = do g <- aux n l f
                                           return $ Ap g (a :| y)

hole :: Type a -> a
hole (PF _) = HOLE

puthole :: Location -> Dynamic -> Dynamic
puthole l (Dyn t x) = Dyn t (xua l (t :| x))
    where xua :: Location -> Typed a -> a
          xua [0] (a :| y) = hole a
	  xua l   (a :| y) = fromSpine $ aux (last l) (init l) (toSpine a y)
	  aux :: Int -> Location -> Spine a -> Spine a
          aux 0     l (Ap f (a :| y)) = Ap f (a :| xua l (a :| y))
          aux (n+1) l (Ap f (a :| y)) = Ap (aux n l f) (a :| y)

-- The basic type of rules
type GenericM m = forall a . Type a -> PF a -> m (PF a)

-- Generic one-level map with location update

gmapMo :: (MonadReader Location m, MonadPlus m) => GenericM m -> GenericM m
gmapMo h t y = do s <- aux h (toSpine (PF t) y)
		  return (fromSpine s)
    where aux :: (MonadReader Location m, MonadPlus m) => GenericM m -> (Spine a -> m (Spine a))
          aux h (c `As` _) = mzero
          aux h (Ap f (PF t :| x)) = (do g <- local next $ aux h f
				         return (Ap g (PF t :| x)))
				     `mplus`
				     (do y <- local down $ h t x
				         return (Ap f (PF t :| y)))

gmapM :: (MonadReader Location m, MonadPlus m) => GenericM m -> GenericM m
gmapM h t y = do s <- aux h (toSpine (PF t) y)
		 return (fromSpine s)
    where aux :: (MonadReader Location m, MonadPlus m) => GenericM m -> (Spine a -> m (Spine a))
          aux h x@(c `As` _) = return x
          aux h (Ap f (PF t :| x)) = do 
            g <- local next $ aux h f
            y <- local down $ h t x
            return (Ap g (PF t :| y))

-- Promoting rules to traversals by updating context and propagating the type
top :: (MonadReader Location m, MonadState Dynamic m, MonadPlus m) => GenericM m -> GenericM m
top f t x = do y <- f t x
	       (Dyn u c) <- get
	       l <- ask
	       let Just z = replace l (PF t :| y) (u :| c) 
	       put (Dyn u z)
	       return y


-- Strategy combinators
once :: (MonadReader Location m, MonadPlus m) => GenericM m -> GenericM m
once f = f ||| gmapMo (once f)

everywhere :: (MonadReader Location m, MonadPlus m) => GenericM m -> GenericM m
everywhere f = f >>> gmapM (everywhere f)

(>>>) :: Monad m => GenericM m -> GenericM m -> GenericM m
(f >>> g) t x = f t x >>= g t

(|||) :: MonadPlus m => GenericM m -> GenericM m -> GenericM m
(f ||| g) t x = f t x `mplus` g t x

many :: MonadPlus m => GenericM m -> GenericM m
many r = (r >>> many r) ||| nop

many1 :: MonadPlus m => GenericM m -> GenericM m
many1 r = r >>> many r

nop :: Monad m => GenericM m
nop t = return

try :: MonadPlus m => GenericM m -> GenericM m
try x = x ||| nop

-- Rewriting monad

type Rewrite = RWST Location [(String,String)] Dynamic Maybe

type Rule = GenericM Rewrite

success :: String -> a -> Rewrite a
success n x = 
    do z <- get
       tell [(n, gshow Dynamic z)]
       return x

context :: Rewrite (Typed Dynamic)
context =
    do l <- ask
       y <- get
       return (Dynamic :| puthole l y)

-- Simplification wrapers

simplify :: Typeable a => Bool -> Rule -> PF a -> IO (PF a)
simplify b s e = do
  Just (x,l) <- return $ evalRWST (s typeof e) [0] (Dyn typeof e)
  when b $ sequence_ (map aux l)
  putStrLn ("  "++(gshow typeof x))
  return x
 where aux (n,y) =putStrLn ("  "++y) >> putStrLn ("=   { "++n++" }") 

reduce :: Rule -> Type a -> (PF a) -> (PF a)
reduce s t x = maybe x fst (evalRWST (s t x) [0] (Dyn (PF t) x))
