{-# OPTIONS -fglasgow-exts #-}

-----------------------------------------------------------------------------
-- |
-- Maintainer  :  joost.visser@di.uminho.pt, alcino@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
-- Point-free variation of TwoLevel.hs
--
-----------------------------------------------------------------------------

module Data.Transform.TwoLevelPF where

import Data.Set as Set
import Data.Map as Map
import Data.List as List
import Control.Monad.State
import Data.Maybe

import Data.Transform.Type
    
-----------------------------------------------------------------------------
-- * Encapsulation of type-changing transformations

-- | Pairs of bi-directional value-level data transformation functions.
data Rep a b = Rep {to :: PF (a -> b), from :: PF (b -> a)}


-- | The View constructor expreses that a type a can be represented by
--   a type b, as witnessed by value-level functions that convert 
--   between these types. Note that the target type does not escape, 
--   which means that it is implicitly existentially qualified.
data View a where
    View :: Rep a b -> Type b -> View (Type a)

instance Show (View a) where
    show (View _ b) = "(View (Rep <to> <from>) "++(show b)++")"

-- | Print the string representation of the target type encapsulated in a view.
showType :: View a -> String
showType (View _ b) = show b

-- | Type-changing rewrite rules.
type Rule = forall a . Type a -> Maybe (View (Type a))


----------------------------------------
-- ** Value-level combinators used to define two-level combinators.

idrep :: Rep a a
idrep = Rep {to = ID, from = ID}

comprep :: Type b -> Rep a b -> Rep b c -> Rep a c
comprep b f g = Rep {from = COMP b (from f) (from g), to = COMP b (to g) (to f)}

prodmaprep :: Rep a b -> Rep c d -> Rep (a,c) (b,d)
prodmaprep f g = Rep {to = PROD (to f) (to g), from = PROD (from f) (from g)}

summaprep :: Rep a b -> Rep c d -> Rep (Either a c) (Either b d)
summaprep f g = Rep {to = SUM (to f) (to g), from = SUM (from f) (from g)}

listmaprep :: Rep a b -> Rep [a] [b]
listmaprep f = Rep {to = LIST (to f), from = LIST (from f)}

setmaprep :: Rep a b -> Rep (Set a) (Set b)
setmaprep f = Rep {to = SET (to f), from = SET (from f)}

mapmaprep :: Rep b c -> Rep (Map a b) (Map a c)
mapmaprep f = Rep {to = MAP (to f), from = MAP (from f)}

mapkeysrep :: Rep a b -> Rep (Map a c) (Map b c)
mapkeysrep f = Rep {to = MAPKEYS (to f), from = MAPKEYS (from f)}

-----------------------------------------------------------------------------
-- * Strategy combinators for two-level transformations

-- | Sequential composition.
(>>>) :: Rule -> Rule -> Rule
(f >>> g) a = do 
  View r b <- f a
  View s c <- g b
  return (View (comprep b r s) c)

-- | Left-biased choice.
(|||) :: Rule -> Rule -> Rule
(f ||| g) x = f x `mplus` g x

-- | Do-nothing operation. 
--   Unit for sequential composition. 
--   Left-zero and right-unit for left-biased choice.
nop :: Rule
nop x = Just (View idrep x)

-- | Repeat until failure.
--   Zero or more repetitions.
many :: Rule -> Rule
many r = (r >>> many r) ||| nop

-- | Apply argument rule exactly once, at arbitrary depth.
once :: Rule -> Rule
once r Int  = r Int
once r Bool = r Bool
once r Char = r Char
once r One  = r One
once r String = r String
once r (List a) = r (List a) `mplus` 
                  (do View s b <- once r a
                      return (View (listmaprep s) (List b)))
once r (Set a) = r (Set a) `mplus` 
                 (do View s b <- once r a
                     return (View (setmaprep s) (Set b)))
once r (Map a b) = r (Map a b) `mplus`
                   (do View s c <- once r a
                       return (View (mapkeysrep s) (Map c b))) `mplus`
                   (do View s c <- once r b
                       return (View (mapmaprep s) (Map a c)))
once r (Either a b) = r (Either a b) `mplus`
                      (do View s c <- once r a
                          return (View (summaprep s idrep) (Either c b))) `mplus`
                      (do View s c <- once r b
                          return (View (summaprep idrep s) (Either a c)))
once r (Prod a b) = r (Prod a b) `mplus`
                    (do View s c <- once r a
                        return (View (prodmaprep s idrep) (Prod c b))) `mplus`
                    (do View s c <- once r b
                        return (View (prodmaprep idrep s) (Prod a c)))
once r (Tag n a) = r (Tag n a) `mplus` 
                   (do View s b <- once r a
                       return (View s (Tag n b)))
once r x = r x


-----------------------------------------------------------------------------
-- * Unleashing composed data migration functions

-- | Apply the forward value-level function masked by a view.
forth :: View (Type a) -> Type b -> a -> Maybe b
forth (View rep tb') tb a = do {Eq <- teq tb tb'; return (eval (to rep) a)}

-- | Apply the backward value-level function masked by a view.
back :: View (Type a) -> Type b -> b -> Maybe a
back (View rep tb') tb b = do {Eq <- teq tb tb'; return (eval (from rep) b)}

-- | Take the representations of forward and backward value-level 
--   functions out of a view.
unView :: View (Type a) -> Type b -> Maybe (PF (a->b), PF (b->a))
unView (View (Rep to fro) tb') tb = do
  Eq <- teq tb tb'
  return (to,fro)

-----------------------------------------------------------------------------
-- * Rules for format evolution

----------------------------------------
-- ** Enrichment and removal

-- | Add a field (to the right).
--   The forward value-level function inserts a 
--   constant value for the new field.
addfield :: Type b -> b -> Rule
addfield b x a = return $ View rep (Prod a b)
    where rep = Rep (SPLIT ID (pnt x)) FST

pnt x = COMP One (PNT x) BANG

-- | Type of generic value-level queries.
type Query b = forall a . Type a -> PF (a -> b)

-- | Add a field.
--   The forward value-level function inserts a field
--   determined by a query on its input value.
addinfo :: Type b -> Query b -> Rule
addinfo b f a = return $ View rep (Prod a b)
  where
    rep = Rep (SPLIT ID (f a)) FST
    
----------------------------------------
-- ** Generalization and restriction

-- | Add alternative.
--   The backward value level function fails if the added 
--   alternative is present.
addalt :: Type b -> Rule
addalt b a = return $ View rep (Either a b)
    where rep = Rep INL (EITHER ID (pnt undefined))

-- | Allow repetition.
--   Maps optional to a list.
allowRep :: Rule
allowRep (Either a One) = return $ View (Rep opt2seq seq2opt) (List a)
   where
     opt2seq = EITHER (wrap a) NIL
     seq2opt = FUN "seq2opt" (\l -> if (List.null l) then Right () else Left (head l)) 

wrap :: Type a -> PF (a -> [a])
wrap a = COMP (Prod a (List a)) CONS (SPLIT ID (COMP One NIL BANG))

----------------------------------------
-- ** Focus on tags

-- | Perform the argument rule inside the given tag.
inside :: String -> Rule -> Rule
inside n r (Tag m a) | n==m = do View r b <- r a
                                 return $ View r (Tag m b)
inside _ _ _ = mzero

-- | Perform the argument rule on the given tag.
atTag :: String -> Rule -> Rule
atTag n r x@(Tag m a) | n==m = r x
atTag _ _ _ = mzero

-----------------------------------------------------------------------------
-- * Rules for data mapping

-- | Flatten a hierarchical data type.
--   Does not include the rule for recursion elimination.

flatten :: Rule
flatten = many (aux >>> many aux >>> prods) >>> many (once assocr)
    where aux = once (listmap ||| mapsum ||| summap ||| mapprodmap ||| setmap ||| prodsum ||| sumone ||| prodone)

{-
-- | Map a hierarchical datatype to a relational type.
toRDB :: Rule
toRDB = many (once fixastable) >>> many (once enum2int) >>> removetags >>> flatten
-}

----------------------------------------
-- ** Elimination and introduction

-- | Eliminate list by refinement to map.
listmap :: Rule
listmap (List a) = return $ View (Rep LIST2MAP ELEMS) (Map Int a)
listmap _ = mzero

-- | Eliminate set by refinement to map.
setmap :: Rule
setmap (Set a) = return $ View (Rep (FUN "set2map" set2fm) (FUN "map2set" dom)) (Map a One) 
  where 
    set2fm s = Map.fromDistinctAscList (zip (Set.toAscList s) (repeat ()))
    dom = Set.fromDistinctAscList . keys
setmap _ = mzero

-- | Eliminate optional by refinement to map.
sumone :: Rule
sumone (Either a One) = return $ View (Rep (FUN "optelim" optelim) (FUN "optintro" optintro)) (Map One a)
 where
   optelim (Left x) = Map.singleton () x
   optelim (Right ()) = Map.empty
   optintro m | Map.member () m = let Just x = Map.lookup () m in Left x
		      | otherwise = Right ()
sumone _ = fail "Not applicable"



----------------------------------------
-- ** Distribution

-- | Distribute map over sum.
mapsum :: Rule
mapsum (Map a (Either b c))
   = return $ View (Rep (FUN "uncojoin" uncojoin) (FUN "cojoin" (cojoin a))) (Prod (Map a b) (Map a c))
  where
    uncojoin :: Map a (Either b c) -> (Map a b, Map a c)
    uncojoin = (  fmap (either id undefined) . Map.fromDistinctAscList >< 
                  fmap (either undefined id) . Map.fromDistinctAscList) .
               List.partition (isLeft . snd) . Map.toAscList
    -- Pre: non duplicate keys
    cojoin :: Type a -> (Map a b, Map a c) -> Map a (Either b c)
    cojoin t = Map.fromDistinctAscList . 
               sortBy (\x y -> gcompare t (fst x) (fst y)) .
               uncurry (++) . 
               (Map.toAscList . fmap Left >< Map.toAscList . fmap Right)
mapsum _ = mzero

-- | Distribute product over sum.
prodsum :: Rule
prodsum (Prod a (Either b c)) 
  = return $ View (Rep (FUN "distr" distr) (FUN "undistr" undistr)) (Either (Prod a b) (Prod a c))
 where
   distr (x, Left y) = Left (x,y)
   distr (x, Right y) = Right (x,y)
   undistr (Left (x,y)) = (x, Left y)
   undistr (Right (x,y)) = (x, Right y)
prodsum (Prod (Either b c) a)
  = return $ View (Rep (FUN "distl" distl) (FUN "undistl" undistl)) (Either (Prod b a) (Prod c a))
 where
          distl (Left y,x) = Left (y,x)
          distl (Right y,x) = Right (y,x)
          undistl (Left (y,x)) = (Left y,x)
          undistl (Right (y,x)) = (Right y,x)
prodsum _ = mzero

summap :: Rule
summap (Map (Either a b) c) 
   = return $ View (Rep (FUN "unpeither" unpeither) (FUN "peither" peither)) (Prod (Map a c) (Map b c))
  where
    unpeither = ( mapKeysMonotonic (either id undefined) . Map.fromDistinctAscList >< 
                  mapKeysMonotonic (either undefined id) . Map.fromDistinctAscList
                ) . List.partition (isLeft . fst) . Map.toAscList
    peither = Map.fromDistinctAscList . uncurry (++) . 
              (fmap (Left >< id) . Map.toAscList >< fmap (Right >< id) . Map.toAscList)
summap _ = mzero





----------------------------------------
-- ** Split, join

-- | Split and join.
mapprodmap :: Rule
mapprodmap (Map a (Prod b (Map c d))) 
  = return $ View mapcomposite (Prod (Map a b) (Map (Prod a c) d))
 where
  -- Precondition to from: all keys a in second table must exist in the first
  -- mapcomposite :: Type a -> Rep (Map a (b, Map c d)) (Map a b, Map (a,c) d)
  mapcomposite = Rep unnjoinProd njoinProd
  unnjoinProd = SPLIT (MAP FST) (COMP (Map a (Map c d)) (FUN "unnjoin" unnjoin) (MAP SND))
  njoinProd = COMP (Prod (Map a b) (Func a (Map c d))) LJOIN (PROD ID (FUN "njoin" njoin))
  unnjoin :: Map a (Map b c) -> Map (a,b) c
  unnjoin = Map.fromDistinctAscList . concat . 
            List.map ((\ (x,l) -> List.map ((const x /\ id) >< id) l) . (id >< Map.toAscList)) . 
            Map.toAscList
  --njoin :: Map (a,b) c -> (a -> (Map b c))
  njoin m x = (Map.fromDistinctAscList . (List.map (snd >< id)) . (List.filter (\ (y,b) -> gcompare a x (fst y) == EQ)) . Map.toAscList) m 
mapprodmap _ = mzero

{-

----------------------------------------
-- ** Recursion

-- | Type of tables.
type Table f = (Map Int (f Int), Int)

-- | Refine fix-point datatype to table type.
fixastable :: Rule
fixastable (Mu f) = do
  (View (Rep t f) fInt) <- applyF f Int
  let rep = Rep {to = (Map.map t >< id) . fixtotable, from = tabletofix . (Map.map f >< id)}
  return (View rep (Prod (Map Int (fInt)) Int)) 
   where
    fixtotable :: Dist f => Mu f -> Table f
    fixtotable mu = execState (mcata aux mu) (Map.empty,-1)
      where 
        aux :: f Int -> State (Table f) Int
        aux i = do
          (t,k) <- get
          put (Map.insert (k+1) i t, k+1)
          return (k+1)
    -- Precondition to abs: non circular and all keys exist
    tabletofix :: Functor f => Table f -> Mu f
    tabletofix = ana aux
      where 
        aux :: Functor f => Table f -> f (Table f)
        aux (t,k) = fmap (\k -> (t,k)) (fromJust $ Map.lookup k t)
fixastable _ = Nothing

-- | Unlift a functor while applying it to a type.
--   Previously: removefunctors.
applyF :: Fctr f -> Type a -> Maybe (View Rep (Type (f a)))
applyF Id x        = return (View (Rep {to = unIdent, from = Ident}) x)
applyF (K a) x     = return (View (Rep {to = unConst, from = Const}) a)
applyF (g:*:h) x   = do
  (View (Rep g2gx gx2g) gx) <- applyF g x
  (View (Rep h2hx hx2h) hx) <- applyF h x
  let rep = Rep {to = \(Pair x y) -> (g2gx x,h2hx y), from = \(x,y) -> Pair (gx2g x) (hx2h y)}
  return (View rep (Prod gx hx))
applyF (g:+:h) x = do
  (View (Rep g2gx gx2g) gx) <- applyF g x
  (View (Rep h2hx hx2h) hx) <- applyF h x
  let rep = Rep {to = (g2gx-|-h2hx) . aux, from = xua . (gx2g-|-hx2h)}
  return (View rep (Either gx hx))
      where 
        aux (Inl x) = Left x
        aux (Inr x) = Right x
        xua (Left x)  = Inl x
        xua (Right x) = Inr x
applyF (g:@:h) x   = do 
  (View (Rep to from) hx) <- applyF h x
  let rep = Rep (fmap to . unComp) (Comp . fmap from)
  return (View rep (g hx))
-}

-----------------------------------------------------------------------------
-- * Associativity and commutativity

-- | Associate product to the right. One step.
assocr :: Rule
assocr (Prod (Prod a b) c) = return $ View (Rep to fro) (Prod a (Prod b c))
  where 
    to  = SPLIT (COMP (Prod a b) FST FST) (PROD SND ID) -- fst.fst /\ snd><id
    fro = SPLIT (PROD ID FST) (COMP (Prod b c) SND SND) -- id><fst /\ snd.snd
assocr _ = mzero

-- | Associate product to the left. One step.
assocl :: Rule
assocl (Prod a (Prod b c)) = return $ View (Rep to fro) (Prod (Prod a b) c)
  where
    to  = SPLIT (PROD ID FST) (COMP (Prod b c) SND SND) -- id><fst /\ snd.snd
    fro = SPLIT (COMP (Prod a b) FST FST) (PROD SND ID) -- fst.fst /\ snd><id 
assocl _ = mzero

-- | Put maps to the right and non-maps to the left.
--   Needs to be redesigned....

prods :: Rule
prods = many (many (once assocr) >>> once swap) >>> many (once assocl)
 where
  swap :: Rule
  swap (Prod a (Map b c)) = mzero
  swap (Prod a (Prod (Map b c) d)) = mzero
  swap (Prod (Map a b) c) = return $ View (Rep to fro) (Prod c (Map a b))
    where
      to  = SPLIT SND FST -- snd/\fst
      fro = SPLIT SND FST -- snd/\fst
  swap _ = mzero

-- | Right unit for product.
prodone :: Rule
prodone (Prod a One) = return $ View (Rep to fro) a
 where
    to  = FST              -- fst
    fro = SPLIT ID BANG    -- \x -> (x,())
prodone _ = mzero

-----------------------------------------------------------------------------
-- * Miscelaneous rules

-- | Refine an n-sum of taged ones to an integer.
--   In other words, interpret the n-sum as an enumeration type.

{-
enum2int :: Rule''
enum2int = (many (once eitherRightAssoc)) >>> e2i 0 where
  e2i :: Int -> Rule''
  e2i i (Tag tg One) = do
    return (View (RepPF (PNT i) (FUN "i2eBase" (ifeq () (error "Enum invalid") i))) Int)
  e2i i a@(Either (Tag tg One) b) = do
    (View (RepPF to fro) int) <- e2i (i+1) b
    Eq <- teq Int int
    let to' = VAR ("e2i") -- EITHER (PNT i) (VAR "to") -- to     ---<<< WHY DOESN'T IT WORK???
    let fro' = VAR ("i2e") -- (ifeq (Left ()) (Right . (eval fro)) i)
    return (View (RepPF to' fro') Int)
    -- return (View (RepPF (EITHER (PNT i) to) (FUN ("i2e"++show i) (ifeq (Left ()) (Right . (eval fro)) i))) b)   
  e2i _ _ = Nothing
ifeq l r i i' = if i==i' then l else r i'
-}

-- | Associate sums to the right.
--   Used by 'enum2int' to discover enumeration types.
eitherRightAssoc :: Rule
eitherRightAssoc (Either (Either a b) c) 
   = return $ View (Rep to from) (Either a (Either b c))
  where
      to = EITHER (SUM ID INL) (COMP (Either b c) INR INR)
      from = EITHER (COMP (Either a b) INL INL) (SUM INR ID) 
eitherRightAssoc _ = mzero

-- | Remove all tags from a type represention. This is useful to
--   allow later rules to assume no tags are present.
removetags :: Rule
removetags = many (once aux)
    where aux (Tag n x) = return $ View idrep x
	  aux _ = mzero


-----------------------------------------------------------------------------
-- * Auxiliary types and functions

-- | Determine whether the left alternative is present.
isLeft :: Either a b -> Bool
isLeft = either (const True) (const False)

----------------------------------------
-- ** Pointfree operators

infix 5 ><
infix 4 /\
infix 3 -|-
infix 2 \/

(f >< g) (x,y) = (f x, g y)
(/\) f g x = (f x, g x)
(f -|- g) (Left x) = Left (f x)
(f -|- g) (Right x) = Right (g x)
(\/) f g (Left x) = f x
(\/) f g (Right x) = g x



-----------------------------------------------------------------------------
-- * Using dynamic types

-- | Apply a generic function to a dynamically typed value.
applyDyn :: (forall a . Type a -> a -> b) -> Dynamic -> b
applyDyn f (Dyn ta a) = f ta a

-- | Turn a function on dynamics into a generic function.
dynApply :: (Dynamic -> b) -> Type a -> a -> b
dynApply f ta a = f (Dyn ta a)

instance Show Dynamic where
    show (Dyn a x) = "Dyn "++(gshow a x)
    
-- | Variation on 'forth' that does not need static knowledge of the
--   target type, and produces a dynamically typed value instead.
forthDyn :: View (Type a) -> a -> Maybe Dynamic
forthDyn (View rep b) x = return (Dyn b (eval (to rep) x))

-- | Variation on 'back' that does not need static knowledge of the
--   original type, and produces a dynamically typed value instead.
backDyn :: View (Type a) -> Dynamic -> Maybe a
backDyn (View rep b) (Dyn c x) = do
  Eq <- teq b c
  return (eval (from rep) x)


-----------------------------------------------------------------------------
-- * Type-indexed functions

-- | Generic show function.
gshow :: Type a -> a -> String
gshow Int i = show i
gshow One _ = "()"
gshow Bool x = show x
gshow Char x = show x
gshow String x = show x
gshow (Prod a b) (x,y) = "("++(gshow a x)++","++(gshow b y)++")"
gshow (Either a b) (Left y) = "Left ("++(gshow a y)++")"
gshow (Either a b) (Right y) = "Right ("++(gshow b y)++")"
gshow (List a) l = "["++(concat $ intersperse "," $ fmap (gshow a) l)++"]"
gshow (Map a b) m = "{"++(concat $ intersperse "," $ fmap aux (Map.toList m))++"}"
    where aux (x,y) = (gshow a x)++":="++(gshow b y)
gshow (Set a) s = "{"++(concat $ intersperse "," $ fmap (gshow a) (Set.elems s))++"}"
gshow (Tag n a) x = n++"="++(gshow a x)
{-
gshow (Mu f) (In x) = maybe "Mu <f>" id aux
  where
    aux :: Maybe String
    aux = do
      (View rep fmf) <- applyF f (Mu f)
      return $ gshow fmf $ (to rep) x
-}


-- | Generic comparison function.
--   Currently defined only for products and integers....
gcompare :: Type a -> a -> a -> Ordering
gcompare Int x y = compare x y
gcompare (Prod a b) (x1,y1) (x2,y2) 
 = let t = gcompare a x1 x2
   in case t of 
     EQ -> gcompare b y1 y2
     otherwise -> t

-----------------------------------------------------------------------------


