{-# OPTIONS -fglasgow-exts #-}
{-# OPTIONS -fallow-undecidable-instances #-}
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  joost.visser@di.uminho.pt, alcino@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
-- Type-safe representation of types at the value level, including
-- representation of recursive types as fixpoints of functors.
--
-- See
--
--     Type-safe Two-level Data Transformation
-- 
-- by
--
--     Alcino Cunha, Jose Nuno Oliveira, Joost Visser
--
-----------------------------------------------------------------------------

module Data.Transform.Type where

import Data.Set as Set
import Data.Map as Map
import Control.Monad

-----------------------------------------------------------------------------
-- * Representation of Types

-- | A type-safe representation of types.
data Type a where

    -- Without fixpoints
    Int :: Type Int
    Bool :: Type Bool
    Char :: Type Char
    String :: Type String
    One :: Type One
    List :: Type a -> Type [a]
    Set :: Type a -> Type (Set a)
    Map :: Type a -> Type b -> Type (Map a b)
    Either :: Type a -> Type b -> Type (Either a b)
    Prod :: Type a -> Type b -> Type (a,b)
    Func :: Type a -> Type b -> Type (a -> b)
    Tag :: String -> Type a -> Type a

    -- Fixpoints
    Mu :: Dist f => Fctr f -> Type (Mu f)
    
    PF :: Type a -> Type (PF a)
    Dynamic :: Type Dynamic

-- | A type-safe representation of lifted functors.
data Fctr f where
    Id    :: Fctr Id
    K     :: Type a -> Fctr (K a)
    (:+:) :: Fctr g -> Fctr h -> Fctr (g:+:h)
    (:*:) :: Fctr g -> Fctr h -> Fctr (g:*:h)
    (:@:) :: Functor g => (forall a . Type a -> Type (g a)) -> Fctr h -> Fctr (g:@:h)

instance Show (Type a) where
    show Int = "Int"
    show Bool = "Bool"
    show Char = "Char"
    show String = "String"
    show One = "One"
    show (List x) = "(List "++(show x)++")"
    show (Set x) = "(Set "++(show x)++")"
    show (Map x y) = "(Map "++(show x)++" "++(show y)++")"
    show (Either x y) = "(Either "++(show x)++" "++(show y)++")"
    show (Prod x y) = "(Prod "++(show x)++" "++(show y)++")"
    show (Func x y) = "("++show x++" -> "++show y++")"
    show (Tag n t) = "(Tag "++(show n)++" "++(show t)++")"

    show (PF a) = "(PF "++show a++")"
    show (Dynamic) = "Dynamic"
    show (Mu f) = "(Mu "++show f++")"
    
instance Show (Fctr f) where
  show Id = "Id"
  show (K t) = "(K "++show t++")"
  show (f:*:g) = "("++show f++":*:"++show g++")"
  show (f:+:g) = "("++show f++":+:"++show g++")"
  show (f:@:g) = "(<?>:@:"++show g++")"
 
----------------------------------------
-- ** Type One

-- data One
type One = ()

_L :: One
_L = error "Attempt to evaluate _L"

-- instance Show One where
--     show _ = "_L"

----------------------------------------
-- ** Type Dynamic

-- | Dynamic types
data Dynamic where
    Dyn :: Type a -> a -> Dynamic

-----------------------------------------------------------------------------
-- * Deriving type representations from types.

-- | Typeable class of types for which we can derive representations.
class Typeable a where
    typeof :: Type a

instance Typeable Int where
    typeof = Int

instance Typeable () where
    typeof = One

instance Typeable Bool where
    typeof = Bool

instance Typeable Char where
    typeof = Char

instance (Typeable a, Typeable b) => Typeable (a,b) where
    typeof = Prod typeof typeof

instance (Typeable a, Typeable b) => Typeable (Either a b) where
    typeof = Either typeof typeof

instance (Typeable a, Typeable b) => Typeable (Map a b) where
    typeof = Map typeof typeof

instance (Typeable a, Typeable b) => Typeable (a -> b) where
    typeof = Func typeof typeof

instance Typeable a => Typeable (PF a) where
    typeof = PF typeof

instance Typeable Dynamic where
    typeof = Dynamic

instance Typeable String where
    typeof = String

instance Typeable a => Typeable [a] where
    typeof = List typeof

instance Typeable a => Typeable (Set a) where
    typeof = Set typeof


-----------------------------------------------------------------------------
-- * Type-indexed functions

----------------------------------------
-- ** Some generic functions

-- | Explicitly contrain a value to a particular type.
constrain :: Type a -> a -> a
constrain _ a = a

-- | Explicitly coerce a value of a given type to another given type.
coerce :: MonadPlus m => Type a -> Type b -> a -> m b
coerce a b x = do Eq <- teq a b
                  return x

----------------------------------------
-- ** Equality of types

-- | Data type to express type-equality.    
data Equal a b where
    Eq :: Equal a a
    
-- | Test equality of types.
--   If the types are equal, a witness 'Eq' of the equality is returned.
teq :: MonadPlus m => Type a -> Type b -> m (Equal a b)
teq Dynamic Dynamic = return Eq
teq Int Int = return Eq
teq Bool Bool = return Eq
teq Char Char = return Eq
teq One One = return Eq
teq String String = return Eq
teq (List a) (List b) = 
    do Eq <- teq a b
       return Eq
teq (Set a) (Set b) = 
    do Eq <- teq a b
       return Eq
teq (Map a b) (Map c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (Prod a b) (Prod c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (Either a b) (Either c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (Func a b) (Func c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (PF a) (PF b) = 
    do Eq <- teq a b
       return Eq
teq (Tag _ a) b =
    do Eq <- teq a b
       return Eq
teq a (Tag _ b) =
    do Eq <- teq a b
       return Eq
teq (Mu f) (Mu g) = do
  error "Cannot determine equality for Mu's"
teq _ _ = mzero


-----------------------------------------------------------------------------
-- * Representation of recursive datatypes.

----------------------------------------
-- ** Standard functors

infixr 5 :+:
infixr 6 :*:
infixr 9 :@:

newtype Id a        = Ident {unIdent :: a}       deriving Eq
newtype K b a       = Const {unConst :: b}       deriving Eq
data (g :+: h) a    = Inl (g a) | Inr (h a)      deriving Eq
data (g :*: h) a    = Pair (g a) (h a)           deriving Eq
newtype (g :@: h) a = Comp {unComp :: g (h a)}   deriving Eq

instance Functor Id where
    fmap f = Ident . f . unIdent

instance Functor (K a) where
    fmap f = Const . id . unConst

instance (Functor f, Functor g) => Functor (f :+: g) where
    fmap f (Inl x) = Inl (fmap f x)
    fmap f (Inr x) = Inr (fmap f x)

instance (Functor f, Functor g) => Functor (f :*: g) where
    fmap f (Pair x y) = Pair (fmap f x) (fmap f y)

instance (Functor g, Functor h) => Functor (g :@: h)
    where fmap f = Comp . fmap (fmap f) . unComp

----------------------------------------
-- ** Fixpoints and recursion patterns

-- | Explicit fixpoint operator
newtype Mu f = In {out :: f (Mu f)}

-- This instance requires -fallow-undecidable-instances.
instance Eq (f (Mu f)) => Eq (Mu f) where
  (In fmf) == (In gmg) = fmf==gmg

-- | Catamorphism or fold
cata :: Functor f => (f a -> a) -> Mu f -> a
cata g = g . fmap (cata g) . out

-- | Anamorphism or unfold
ana :: Functor f => (a -> f a) -> a -> Mu f
ana g = In . fmap (ana g) . g

-- | Natural transformation
nu :: Functor g => (forall a. f a -> g a) -> Mu f -> Mu g
nu t = In . fmap (nu t) . t . out

----------------------------------------
-- ** Functors and monads

-- | Class of functors that distribute over monads.
class Functor f => Dist f where
    dist :: Monad m => f (m a) -> m (f a)

instance Dist Id where
    dist (Ident x) = do y <- x
			return (Ident y)

instance Dist (K a) where
    dist (Const x) = return (Const x)

instance (Dist f, Dist g) => Dist (f :+: g) where
    dist (Inl x) = do y <- dist x
		      return (Inl y)
    dist (Inr x) = do y <- dist x
		      return (Inr y)

instance (Dist f, Dist g) => Dist (f :*: g) where
    dist (Pair l r) = do x <- dist l
			 y <- dist r
			 return (Pair x y)

instance (Dist g, Dist h) => Dist (g :@: h) where
    dist (Comp x) = do y <- return $ fmap dist x
		       z <- dist y
		       return (Comp z)

instance Dist [] where
    dist = sequence


-- | Monadic catamorphism.
mcata :: (Monad m, Dist f) => (f a -> m a) -> Mu f -> m a
mcata g = cata (\x -> dist x >>= g)

-----------------------------------------------------------------------------
-- * Type-safe representation of point-free functions.

-- | GADT to represent point-free functions.
data PF a where
    HOLE :: PF a
    FUN :: String -> (a -> b) -> PF (a -> b)
    BANG :: PF (a -> One)
    PNT :: a -> PF (One -> a)
    ID :: PF (a -> a)
    COMP :: Type b -> PF (b -> c) -> PF (a -> b) -> PF (a -> c)
    FST :: PF ((a,b) -> a)
    SND :: PF ((a,b) -> b)
    SPLIT :: PF (a -> b) -> PF (a -> c) -> PF (a -> (b,c))
    PROD :: PF (a -> b) -> PF (c -> d) -> PF ((a,c) -> (b,d))
    INL :: PF (a -> Either a b)
    INR :: PF (b -> Either a b)
    EITHER :: PF (a -> c) -> PF (b -> c) -> PF (Either a b -> c)
    SUM :: PF (a -> b) -> PF (c -> d) -> PF (Either a c -> Either b d)
    AP :: PF ((a -> b, a) -> b)
    CURRY :: PF ((a,b) -> c) -> PF (a -> (b -> c))
    EXP :: PF (a -> b) -> PF ((c -> a) -> (c -> b))
    NIL :: PF (One -> [a])
    CONS :: PF ((a,[a]) -> [a])
    LIST :: PF (a -> b) -> PF ([a] -> [b])
    LIST2MAP :: PF ([a] -> Map Int a)
    EMPTY :: PF (One -> Map a b)
    INSERT :: PF (((a,b),Map a b) -> Map a b)
    MAP :: PF (b -> c) -> PF (Map a b -> Map a c)
    MAPKEYS :: PF (a -> b) -> PF (Map a c -> Map b c)
    ELEMS :: PF (Map a b -> [b])
    LJOIN :: PF ((Map a b, a -> c) -> Map a (b,c))
    SET :: PF (a -> b) -> PF (Set a -> Set b)

instance Show (PF (a->b)) where
  show (FUN s _) = s
  show BANG = "bang"
  show (PNT a ) = "(PNT"++"_)"       
  show ID  = "id"
  show (COMP t f g) = "("++show f++" . "++show g++")"
  show FST  = "fst"
  show SND  = "snd"
  show (SPLIT f g) = "("++show f++" /\\ "++show g++")"
  show (PROD f g) = "("++show f++" >< "++show g++")"
  show INL  = "inl"
  show INR  = "inr"
  show (EITHER f g) = "("++show f++" \\/ "++show g++")"
  show (SUM f g) = "("++show f++" -|- "++show g++")"
  show AP = "ap"
  show (CURRY f) = "(curry"++show f++")"
  show (EXP f) = "(exp"++show f++")"
  show NIL   = "nil"
  show CONS  = "cons"
  show (LIST f) = "(List.map "++show f++")"
  show (LIST2MAP) = "list2map"
  show EMPTY = "empty"
  show INSERT = "insert"
  show (MAP f) = "(Map.map "++show f++")"
  show (MAPKEYS f) = "(Map.mapKeys "++show f++")"
  show ELEMS = "elems"
  show LJOIN = "ljoin"
  show (SET f) = "(Set.map "++show f++")"


-- Evaluation of point-free terms VERY INCOMPLETE!

eval :: PF a -> a
eval (FUN n f) = f
eval BANG = const _L
eval (PNT x) = const x
eval ID = id
eval (COMP _ f g) = eval f . eval g
eval FST = fst
eval SND = snd
eval (SPLIT f g) = \x -> (eval f x, eval g x)
eval (PROD f g) = \(x,y) -> (eval f x, eval g y)
eval INL = Left
eval INR = Right
eval (EITHER f g) = either (eval f) (eval g)
eval (SUM f g) = either (Left . eval f) (Right . eval g)
eval AP = uncurry ($)
eval (CURRY f) = curry (eval f)
eval (EXP f) = (.) (eval f) 
eval NIL = const []
eval CONS = uncurry (:)
eval (LIST f) = Prelude.map (eval f)
eval LIST2MAP = Map.fromDistinctAscList . (zip [0..])
eval EMPTY = const Map.empty
eval INSERT = undefined
eval (MAP f) = Map.map (eval f)
eval (MAPKEYS f) = Map.mapKeysMonotonic (eval f) -- Wrong!!
eval ELEMS = Map.elems
eval LJOIN = undefined
eval (SET f) = Set.mapMonotonic (eval f) -- Wrong!!

-----------------------------------------------------------------------------
